import os


def to_bool(s):
    if s is None:
        return None
    if isinstance(s, bool):
        return s
    if isinstance(s, str) and len(s):
        if s.lower() in ("f", "false", "n", "no", "0"):
            return False
        return True
    return False


def get_bool_from_env(name, default):
    return to_bool(os.environ.get(name, default))


DEBUG = get_bool_from_env("DEBUG", False)
COLORS = get_bool_from_env("COLORS", None)



import math

try:
    from pythonjsonlogger import jsonlogger
except ImportError as exc:
    raise ImportError('To use "{}" you must install "python-json-logger" package.'.format(__name__)) from exc


class StackdriverJsonFormatter(jsonlogger.JsonFormatter):
    def __init__(self, fmt="%(levelname) %(message)", style="%", *args, **kwargs):
        super().__init__(fmt=fmt, *args, **kwargs)

    def process_log_record(self, log_record):
        log_record["severity"] = log_record["levelname"]
        del log_record["levelname"]
        return super(StackdriverJsonFormatter, self).process_log_record(log_record)

    def add_fields(self, log_record, record, message_dict):
        subsecond, second = math.modf(record.created)
        log_record.update(
            {
                "timestamp": {"seconds": int(second), "nanos": int(subsecond * 1e9)},
                "thread": record.thread,
                "severity": record.levelname,
            }
        )
        super().add_fields(log_record, record, message_dict)



import codecs
import logging
import sys

from mondrian import errors, filters, formatters, levels, term, humanizer, settings
from mondrian._version import __version__

__all__ = [
    "__version__",
    "humanizer",
    "errors",
    "filters",
    "formatters",
    "levels",
    "settings",
    "setup",
    "setup_excepthook",
    "term",
]

# Patch standard output/error if it's not supporting unicode
# See: https://stackoverflow.com/questions/27347772/print-unicode-string-in-python-regardless-of-environment
if sys.stdout.encoding is None or sys.stdout.encoding == "ANSI_X3.4-1968":
    sys.stdout = codecs.getwriter("UTF-8")(sys.stdout.buffer, errors="replace")
    sys.stderr = codecs.getwriter("UTF-8")(sys.stderr.buffer, errors="replace")


def setup_excepthook():
    """
    Replace default python exception hook with our own.

    """
    sys.excepthook = errors.excepthook


is_setup = False


def setup(*, level=None, colors=term.usecolors, excepthook=False, formatter=None):
    """
    Setup mondrian log handlers.

    :param colors: bool
    :param excepthook: bool
    """
    global is_setup

    logger = logging.getLogger()

    if not is_setup:
        handler = logging.StreamHandler(sys.stderr)

        if formatter:
            handler.setFormatter(formatter)
        else:
            for _level, _name in levels.NAMES.items():
                logging.addLevelName(_level, _name)
            handler.setFormatter(formatters.Formatter())

        handler.addFilter(filters.ColorFilter() if colors else filters.Filter())
        logger.addHandler(handler)
        logging.captureWarnings(True)

        is_setup = True

    if excepthook:
        setup_excepthook()

    if level is None:
        # set default based on env
        logger.setLevel(logging.DEBUG if settings.DEBUG else logging.INFO)
    elif level is not False:
        logger.setLevel(level)


def getLogger(*args, colors=(not term.iswindows), excepthook=False, **kwargs):
    setup(colors=colors, excepthook=excepthook)
    return logging.getLogger(*args, **kwargs)



import itertools
import re
import sys
from contextlib import contextmanager
from sys import exc_info

from mondrian import settings, term
from mondrian.formatters import format_exception
from mondrian.styles import BOTTOM_LEFT, BOTTOM_RIGHT, HORIZ, TOP_LEFT, TOP_RIGHT, VERT, VERT_LEFT

preformatted_pattern = re.compile("([^`]*)`([^`]*)`([^`]*)")


def humanized(exc, *, fg=term.red, bg=lambda *args: term.red_bg(term.bold(*args)), help_url=None):
    SPACES = 2
    prefix, suffix = fg(VERT + " " * (SPACES - 1)), fg(" " * (SPACES - 1) + VERT)
    result = []

    def format_arg(arg):
        length = len(preformatted_pattern.sub("\\1\\2\\3", arg))
        arg = preformatted_pattern.sub("\\1" + term.bold("\\2") + "\\3", arg)
        arg = re.sub("^  \$ (.*)", term.lightblack("  $ ") + term.reset("\\1"), arg)
        return (arg, length)

    def joined(*args):
        return "".join(args)

    term_width, term_height = term.get_size()
    line_length = min(80, term_width)
    for arg in exc.args:
        line_length = max(min(line_length, len(arg) + 2 * SPACES), 120)

    result.append(joined(fg(TOP_LEFT + HORIZ * (line_length - 2) + TOP_RIGHT)))

    args = list(exc.args)

    for i, arg in enumerate(args):

        if i == 1:
            result.append(joined(prefix, " " * (line_length - 2 * SPACES), suffix))

        arg_formatted, arg_length = format_arg(arg)
        if not i:
            # first line
            result.append(
                joined(
                    prefix,
                    bg(" " + type(exc).__name__ + " "),
                    " ",
                    term.white(arg_formatted),
                    " " * (line_length - (arg_length + 3 + len(type(exc).__name__) + 2 * SPACES)),
                    suffix,
                )
            )
        else:
            # other lines
            result.append(joined(prefix, arg_formatted + " " * (line_length - arg_length - 2 * SPACES), suffix))

    if help_url:
        help_prefix = 'Read more: '
        arg_length = len(help_url) + len(help_prefix)
        arg_formatted = help_prefix + term.underline(term.lightblue(help_url))
        result.append(joined(prefix, " " * (line_length - 2 * SPACES)  , suffix))
        result.append(joined(prefix, arg_formatted+" " * (line_length - arg_length - 2 * SPACES)  , suffix))


    more = settings.DEBUG
    exc_lines = format_exception(exc_info(), fg=fg, bg=bg, summary=False).splitlines()

    if not len(exc_lines):
        more = False

    result.append(
        joined(fg((VERT_LEFT if more else BOTTOM_LEFT) + HORIZ * (line_length - 2) + BOTTOM_RIGHT)),
    )

    if more:
        for _line in exc_lines:
            result.append(_line)
        result.append(joined(fg("╵")))

    return '\n'.join(result)


class Success():
    def __init__(self, *args, help_url=None):
        self.args = args
        self.help_url = help_url

    def __str__(self):
        return humanized(self, fg=term.green, bg=(lambda *args: term.lightgreen_bg(term.lightblack(*args))), help_url=self.help_url)


@contextmanager
def humanize(*, types=(Exception,)):
    if not isinstance(types, tuple):
        types = (types,)

    try:
        yield
    except types as exc:
        print(humanized(exc), file=sys.stderr)



import logging
import re
import textwrap
from traceback import format_exception as _format_exception

from colorama import Style

from mondrian import settings, styles
from mondrian.term import CLEAR_EOL, iswindows, lightblack, lightblack_bg, lightwhite

if iswindows or not settings.COLORS:
    EOL = "\n"
else:
    EOL = Style.RESET_ALL + CLEAR_EOL + "\n"


class _Style(logging.PercentStyle):
    def __init__(self):
        tokens = ("%(color)s%(levelname)s", "%(spent)04d", "%(name)s")

        sep = "|" if iswindows else (Style.RESET_ALL + lightblack(":"))
        self._fmt = sep.join(tokens) + lightblack(":") + " %(message)s" + EOL.strip()


def format_exception(excinfo, *, prefix="", fg=lightblack, bg=lightblack_bg, summary=True):
    formatted_exception = _format_exception(*excinfo)

    output = []
    stack_length = len(formatted_exception)
    for i, frame in enumerate(formatted_exception):
        _last = i + 1 == stack_length
        if frame.startswith("  "):
            output.append(textwrap.indent("  " + frame.strip(), fg("\u2502 ")))
        else:
            # XXX TODO change this to use valid python package regexp (plus dot).
            g = re.match("([a-zA-Z_.]+): (.*)$", frame.strip(), flags=re.DOTALL)
            if summary or not _last:
                if g is not None:
                    etyp, emsg = g.group(1), g.group(2)
                    output.append(
                        fg(styles.BOTTOM_LEFT if _last else styles.VERT_LEFT)
                        + bg(lightwhite(" " + etyp + " "))
                        + " "
                        + lightwhite(textwrap.indent(str(emsg), " " * (len(etyp) + 4)).strip())
                    )
                else:
                    output.append(textwrap.indent(frame.strip(), fg("\u2502 ")))

    return textwrap.indent(EOL.join(output), prefix=prefix)


class Formatter(logging.Formatter):
    def __init__(self):
        self._style = _Style()
        self._fmt = self._style._fmt

    formatException = staticmethod(format_exception)



import logging

from colorama import Fore

DEBUG = "DEBG"
INFO = "INFO"
WARNING = "WARN"
ERROR = "ERR."
CRITICAL = "CRIT"

COLORS = {
    DEBUG: Fore.LIGHTCYAN_EX,
    INFO: Fore.LIGHTWHITE_EX,
    WARNING: Fore.LIGHTYELLOW_EX,
    ERROR: Fore.LIGHTRED_EX,
    CRITICAL: Fore.RED,
}

NAMES = {
    logging.DEBUG: DEBUG,
    logging.INFO: INFO,
    logging.WARNING: WARNING,
    logging.ERROR: ERROR,
    logging.CRITICAL: CRITICAL,
}
